#  Minder: Simple Calendaring System
#  Copyright (C) 2015 Adam Hartz
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Lexer

import re
import string
import random

from collections import OrderedDict
from datetime import date, time, datetime, timedelta

from minder.util import (days_in_month,
                         MinderError,
                         MinderDateError,
                         MinderTimeError)


class MinderSyntaxError(MinderError):
    pass


def _anyof(*res):
    return '(' + '|'.join(res) + ')'


# Regular Expressions for matching each token type


TOKENS = OrderedDict()

TD = (r'(\d+(?:\.\d*)?)([mhdwMy]|min(?:ute)?s?|h(?:ou)?rs?'
      r'|days?|w(?:ee)?ks?|mo(?:nth|n)?s?|y(?:ea)?rs?)$')
TOKENS['RELTIME'] = re.compile(r'([+-])%s' % TD)
TOKENS['TDELTA'] = re.compile(TD)

TIME = r'(\d{1,2}:\d{2}$)|(\d{1,2}(?::\d{2})?)([ap]m?$)'
TOKENS['TIME'] = re.compile(TIME, re.IGNORECASE)

DATE = r'\d{4}-\d{1,2}-\d{1,2}$'
TOKENS['DATE'] = re.compile(DATE)

WEEKDAYS = {
    'MON': 0, 'MONDAY': 0,
    'TUE': 1, 'TUES': 1, 'TUESDAY': 1,
    'WED': 2, 'WEDNESDAY': 2,
    'THU': 3, 'THUR': 3, 'THURS': 3, 'THURSDAY': 3,
    'FRI': 4, 'FRIDAY': 4,
    'SAT': 5, 'SATURDAY': 5,
    'SUN': 6, 'SUNDAY': 6,
}
TOKENS['WEEKDAY'] = re.compile(_anyof(*WEEKDAYS) + '$', re.IGNORECASE)

MONTHS = {
    'JAN': 1, 'JANUARY': 1,
    'FEB': 2, 'FEBRUARY': 2,
    'MAR': 3, 'MARCH': 3,
    'APR': 4, 'APRIL': 4,
    'MAY': 5,
    'JUN': 6, 'JUNE': 6,
    'JUL': 7, 'JULY': 7,
    'AUG': 8, 'AUGUST': 8,
    'SEP': 9, 'SEPT': 9, 'SEPTEMBER': 9,
    'OCT': 10, 'OCTOBER': 10,
    'NOV': 11, 'NOVEMBER': 11,
    'DEC': 12, 'DECEMBER': 12
}
TOKENS['MONTH'] = re.compile(_anyof(*MONTHS) + '$', re.IGNORECASE)
TOKENS['MONTHDAY'] = re.compile('\d{1,2}/\d{1,2}$')
TOKENS['MONTHDAYYEAR'] = re.compile('\d{1,2}/\d{1,2}/(?:\d{2}|\d{4})$')

TOKENS['YEAR'] = re.compile(r"\d{4}$|'\d{2}$")
TOKENS['DAY'] = re.compile(r'\d{1,2}$')

# Names, classes, events, and options

TOKENS['CLASS'] = re.compile(r'(CLASS|@)$')
TOKENS['OPTION'] = re.compile(r'-$')
TOKENS['EVENT_OR_GROUP'] = re.compile(r'\*+$')
TOKENS['MACRO'] = re.compile(r'(MACRO|=)$')

RULE = _anyof('omit', 'end|until', 'dur(?:ation)?', 'msg|message', 'skip',
              'subj(?:ect)?', 'rem(?:ind)?', 'rep(?:lace|l)?', 'format',
              'inc(?:lude|l)?', 'dest(?:ination)?|recp|recipient|sendto',
              'beg(?:in)?|start|from', 'alert') + '$'
KEYWORD = _anyof('to', 'with') + '$'
TOKENS['RULE'] = re.compile(RULE, re.IGNORECASE)
TOKENS['KEYWORD'] = re.compile(KEYWORD, re.IGNORECASE)

NAME = r'[A-Za-z]\w*$'
TOKENS['CLASSREF'] = re.compile(r'@%s' % NAME)
TOKENS['MACROREF'] = re.compile(r'=%s' % NAME)
TOKENS['NAME'] = re.compile(NAME)

# String handling

TRIPLE_STRING = r'"""((\\(.|\n))|([^"\\])|("(?!""))|\n)*"""'
DOUBLE_STRING = r'"((\\(.|\n))|([^"\\]))*"'
SQUIGGLE_STRING = r'{((\\(.|\n))|([^}\\])|\n)*}'
STRING_LITERAL = _anyof(TRIPLE_STRING, DOUBLE_STRING, SQUIGGLE_STRING)
STRING = re.compile(STRING_LITERAL)


def get_sep(s):
    sep = None
    while sep is None or sep in s:
        sep = ''.join(random.choice(string.ascii_letters) for _ in range(10))
    return sep


def remove_strings_and_comments(s):
    sep = get_sep(s)
    matches = []

    def sep_matcher(m):
        matches.append(m.group(0))
        return sep

    s = re.sub(STRING, sep_matcher, s)
    s = re.sub(r'#.*\n', '', s)
    return s.split(sep), matches


# Support for COND token


def remove_cond(x):
    sep = get_sep(x)
    c = 'COND('
    lx = len(x)
    o = []
    while c in x:
        oix = x.find(c)
        ix = oix + 5
        count = 1
        while ix < lx:
            if x[ix] == '(':
                count += 1
            elif x[ix] == ')':
                count -= 1
            if count == 0:
                break
            ix += 1
        else:
            raise MinderSyntaxError("EOF while scanning COND")
        o.append(x[oix:ix+1])
        x = x[:oix] + sep + x[ix+1:]
    return x.split(sep), o


# Handlers to determine the right value to associate with each token


def get_token_value(name, match):
    return VALUE_FUNCTIONS.get(name, pass_value)(match)


def pass_value(m):
    return m.group(0)


def afterone_value(m):
    return m.group(0)[1:]


def day_value(m):
    return {'day': {int(m.group(0))}}


def year_value(m):
    t = m.group(0)
    if t.startswith("'"):
        return {'year': {2000 + int(t[1:])}}
    return {'year': {int(t)}}


def upper_value(m):
    return m.group(0).upper()


def weekday_value(m):
    return {'dow': {WEEKDAYS[m.group(0).upper()]}}


def month_value(m):
    return {'month': {MONTHS[m.group(0).upper()]}}


def string_value(m):
    if m.startswith('"') or m.startswith("'"):
        return eval(m)
    else:
        return m[1:-1].replace(r'\}', '}')


def reltime_value(m):
    g = m.groups()
    t = tdelta_value_helper(g[1:])
    t['neg'] = g[0] == '-'
    return t


def tdelta_value(m):
    return tdelta_value_helper(m.groups())


def tdelta_value_helper(g):
    num = float(g[0])
    t = g[1]
    out = {'neg': False}
    if t in ('m', 'min', 'mins', 'minute', 'minutes'):
        out['minutes'] = num
    elif t.startswith('h'):
        out['hours'] = num
    elif t.startswith('d'):
        out['days'] = num
    elif t.startswith('w'):
        out['days'] = 7 * num
    elif t in ('M', 'mon', 'mons', 'month', 'months'):
        out['months'] = num
    elif t.startswith('y'):
        out['years'] = num
    return out


def date_value(m):
    return date_value_helper(m.group(0))


def date_value_helper(o):
    y, m, d = map(int, o.split('-'))
    if y > 9999 or y < 1000:
        e = 'Invalid year %d in DATE: %s' % (y, o)
        raise MinderDateError(e)
    if not (1 <= m <= 12):
        e = 'Invalid month %s in DATE: %s' % (m, o)
        raise MinderDateError(e)
    dim = days_in_month(y, m)
    if not (1 <= d <= dim):
        e = 'Invalid day %s in DATE: %s (only %s days in month)' % (d, o, dim)
        raise MinderDateError(e)
    return {'year': {y}, 'month': {m}, 'day': {d}}


def time_value(m):
    return time_value_helper(m.groups(), m.group(0))


def time_value_helper(g, o):
    if g[0] is None:
        h, m = plain_time_helper(g[1])
        first = g[2].upper()[0]
        if first == 'P' and h != 12:
            h += 12
        elif first == 'A' and h == 12:
            h = 0
    else:
        h, m = plain_time_helper(g[0])
    if not (0 <= h < 24):
        raise MinderTimeError('Invalid hour %s in TIME: %s' % (h, o))
    if not (0 <= m < 60):
        raise MinderTimeError('Invalid minute %s in TIME: %s' % (m, o))
    return {'time': {(h, m)}}


def plain_time_helper(t):
    if ':' in t:
        return tuple(map(int, t.split(':')))
    else:
        return (int(t), 0)


def monthday_value(m):
    m, d = tuple(map(int, m.group(0).split('/')))
    return {'month': {m}, 'day': {d}}


def monthdayyear_value(m):
    m, d, y = tuple(map(int, m.group(0).split('/')))
    if y < 100:
        y = 2000
    return date_value_helper('%d-%02d-%02d' % (y, m, d))


VALUE_FUNCTIONS = {
    'DATE': date_value,
    'TIME': time_value,
    'RELTIME': reltime_value,
    'TDELTA': tdelta_value,
    'WEEKDAY': weekday_value,
    'MONTH': month_value,
    'YEAR': year_value,
    'DAY': day_value,
    'MONTHDAY': monthday_value,
    'MONTHDAYYEAR': monthdayyear_value,
    'CLASSREF': afterone_value,
    'MACROREF': afterone_value,
    'RULE': upper_value,
    'KEYWORD': upper_value
}


# Main Tokenization Function


def tokenize(s):
    s, m = remove_strings_and_comments(s)
    num_strings = len(m)
    for ix, piece in enumerate(s):
        s2, m2 = remove_cond(piece)
        for jx, piece2 in enumerate(s2):
            for word in piece2.split():
                found = False
                for (name, r) in TOKENS.items():
                    match = r.match(word)
                    if match:
                        found = True
                        value = get_token_value(name, match)
                        if name == 'MONTHDAYYEAR':
                            name = 'DATE'
                        yield (name, value)
                        break
                if not found:
                    raise MinderSyntaxError("Invalid token: %s" % word)
            if jx < len(m2):
                yield ('COND', m2[jx][5:-1])
        if ix < num_strings:
            yield ('STRING', string_value(m[ix]))
    yield ('EOF', '', None)
