#  Minder: Simple Calendaring System
#  Copyright (C) 2015 Adam Hartz
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Utilities for sending e-mails

import io
import os
import stat
from smtplib import SMTP
from datetime import datetime
from configparser import ConfigParser

from minder.util import MinderError, CONFIG_LOCATION

DEFAULT_MESSAGE = "Reminder: {NAME} on {HUMAN_TIME}"
DEFAULT_FORMAT = "%A %B %d, %Y at %I:%M %p"
DEFAULT_DATE_FORMAT = "%A %B %d, %Y"
DEFAULT_TIME_FORMAT = "%I:%M%p"


class DictConfigParser(ConfigParser):
    def as_dict(self):
        d = dict(self._sections)
        for k in d:
            d[k] = dict(self._defaults, **d[k])
            d[k].pop('__name__', None)
        return d


class MinderEmailError(MinderError):
    pass


def make_summaries(config, date, trigs):
    from_addr = config['from']

    subject = "Minder: Events for %s" % date.strftime(DEFAULT_DATE_FORMAT)

    events = {}
    for (e, t) in trigs:
        if len(e['SENDTO']) == 0:
            raise MinderEmailError('No recipients!')
        for i in e['SENDTO']:
            events[i] = events.get(i,[]) + [(e,t)]

    out = []
    for recipient in events:
        message = subject + '\n' + '-'*len(subject) + '\n'
        for event in sorted(events[recipient], key=lambda x: x[1]):
            message += '%s: %s\n' % (event[1].strftime(DEFAULT_TIME_FORMAT),
                                   event[0]['NAME'])
        msg = "From: {}\nTo: {}\nSubject: {}\n\n{}".format(from_addr,
                                                           recipient,
                                                           subject,
                                                           message)
        out.append((msg, [recipient]))
    return out


def make_message(config, e):
    e = dict(e)

    if 'from' not in config:
        raise MinderEmailError('"from" not found in config!')

    if len(e['SENDTO']) == 0:
        raise MinderEmailError('No recipients!')

    for (s, default) in {('SUBJECT', DEFAULT_MESSAGE),
                         ('MESSAGE', DEFAULT_MESSAGE),
                         ('FORMAT', DEFAULT_FORMAT)}:
        if s not in e or e[s] is None:
            e[s] = default

    m = e['MESSAGE']
    s = e['SUBJECT']

    e['SENT_TIME'] = datetime.now()

    e['HUMAN_TIME'] = e['TRIGGER_TIME'].strftime(e['FORMAT'])
    e['HUMAN_SENT_TIME'] = e['SENT_TIME'].strftime(e['FORMAT'])

    subject = s.format(**e)
    message = m.format(**e)

    from_addr = config['from']
    to_addrs = ', '.join(e['SENDTO'])

    msg = "From: {}\nTo: {}\nSubject: {}\n\n{}".format(from_addr,
                                                       to_addrs,
                                                       subject,
                                                       message)
    return msg, e['SENDTO']


def send_message(config, msg):
    if 'server' not in config:
        raise MinderEmailError('"{}" not found in config!'.format(field))

    port = config.get('port', 25)
    with SMTP(config['server'], port) as smtp:
        smtp.set_debuglevel(False)
        smtp.ehlo()
        if 'username' in config and 'password' in config:
            if smtp.has_extn('STARTTLS'):
                smtp.starttls()
                smtp.ehlo()
            smtp.login(config['username'], config['password'])

        smtp.sendmail(config['from'], msg[1], msg[0])


def get_email_config():
    email_config = os.path.join(CONFIG_LOCATION, 'email.ini')
    if not os.path.isfile(email_config):
        e = 'No email config file found: {} does not exist.'
        raise MinderError(e.format(email_config))
    perms = os.stat(email_config)[stat.ST_MODE]
    if perms & 0o7077:
        e = 'File {} must have no more than -rwx------ (0700) permissions.'
        raise MinderError(e.format(email_config))
    if not perms & stat.S_IRUSR:
        e = 'File {} must be readable by owner.'
        raise MinderError(e.format(email_config))
    config_text = "[config]\n" + open(email_config).read()
    s = io.StringIO(config_text)
    parser = DictConfigParser(allow_no_value=True)
    parser.readfp(s)
    return parser.as_dict()['config']
