#  Minder: Simple Calendaring System
#  Copyright (C) 2015 Adam Hartz
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Main entry point for minder

import os
import stat
from datetime import datetime, time, MINYEAR
from argparse import ArgumentParser, RawDescriptionHelpFormatter

import minder.util as util
from minder import __version__ as VERSION
from minder.lexer import tokenize
from minder.parser import parse, unify_dates, parse_dt_arg
from minder.mail import (get_email_config, make_message,
                         send_message, make_summaries)
from minder.cache import CONFIG_LOCATION, get_last_run, set_last_run
from minder.trigger import get_triggers_could_need_alert, get_triggers_in_range

_logo = ("     .     .       \n"
         ".-.-...-..-| .-,.-.\n"
         "' ' ''' '`-'-`'-'  ")

parser = ArgumentParser(description=_logo+"\nminder: a simple calendaring system\nversion %s" % VERSION,
                        formatter_class=RawDescriptionHelpFormatter)
parser.add_argument('-c',
                    help="Update the cache and exit",
                    dest='only_update_cache',
                    action='store_true')
parser.add_argument('-n',
                    help='Do not send email; rather, print messages to console',
                    dest='mail',
                    action='store_false')
parser.add_argument('-s',
                    help="Send a summary of a day's events",
                    action='store_true',
                    dest='summary')
parser.add_argument('-d',
                    help="The date to consider as the current date",
                    dest='today',
                    default=None)
parser.add_argument('-l',
                    help='The date to consider as the "last run" date',
                    dest='last_run',
                    default=None)
parser.add_argument('-r',
                    help=("Recursively search provided directories "
                          "for *.mnd calendars"),
                    dest='recurse',
                    action='store_true')
parser.add_argument('file',
                    help='Files containing minder calendars',
                    nargs='*')


def main(argv=None):
    args = parser.parse_args()

    email_conf = get_email_config()
    update_cache = True

    # determine what calendars we need to read from
    args.file = [os.path.join(CONFIG_LOCATION, 'default.mnd')] + args.file
    args.file = util.all_calendars(args.file, args.recurse)

    # parse all events out of all calendars
    events = []
    for fname in args.file:
        with open(fname) as f:
            events.extend(parse(f.read()))

    # find the right starting point for our search (or the right day
    # for our summary)
    if args.today is None:
        now = datetime.now()
    else:
        now = parse_dt_arg(args.today, '-d', require_time=(not args.summary))
        update_cache = False

    if args.last_run is None:
        starttime = get_last_run() or datetime(MINYEAR, 1, 1, 0, 0)
    else:
        starttime = parse_dt_arg(args.last_run, '-l')

    # "only update cache" mode: update the cache, and do nothing else
    if args.only_update_cache:
        to_send = []
        update_cache = True

    # normal mode: run a search, find all messages that need to be sent
    elif not args.summary:
        triggers = [(e, get_triggers_could_need_alert(e, starttime))
                    for e in events]
        to_send = []
        for (e, trigs) in triggers:
            for t in trigs:
                if t is None:
                    continue
                evt = dict(e)
                evt['TRIGGER_TIME'] = t
                for r in evt['REMIND']:
                    rtime = util.apply_tdelta(t, r)
                    if util.lte(rtime, now) and util.gte(rtime, starttime):
                        to_send.append(make_message(email_conf, evt))

    # summary mode: return a sumary of today's events, or of args.today's events
    else:
        update_cache = False
        starttime = now.date() if isinstance(now, datetime) else now
        start = datetime.combine(starttime, time(0,0))
        end = datetime.combine(starttime, time(23,59))
        all_triggers = []
        for e in events:
            e = dict(e)
            e['REMIND'] = [{'minutes':0, 'neg':False}]
            all_triggers.extend([(e,i) for i in  get_triggers_in_range(e, start, end)])
        to_send = make_summaries(email_conf, start.date(), all_triggers)


    # send messages (or print, depending)
    if args.mail:
        _doit = lambda m: send_message(email_conf, m)
    else:
        _doit = lambda m: print(m[0])
    for s in to_send:
        _doit(s)
    if update_cache:
        set_last_run(now)

if __name__ == '__main__':
    main()
