#  Minder: Simple Calendaring System
#  Copyright (C) 2015 Adam Hartz
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Parse a stream of tokens into a list of triggers

from itertools import chain

from minder.rule import *
from minder.lexer import tokenize
from minder.util import MinderError


class MinderParseError(MinderError):
    pass

SEPS = {'CLASS', 'OPTION', 'EVENT_OR_GROUP', 'MACRO', 'EOF'}


# Parser pass to unify options

RULE_MAP = {
    'SKIP': 'OMIT',
    'UNTIL': 'END',
    'DUR': 'DURATION',
    'MSG': 'MESSAGE',
    'SUBJ': 'SUBJECT',
    'REM': 'REMIND',
    'ALERT': 'REMIND',
    'REP': 'REPLACE',
    'REPL': 'REPLACE',
    'INC': 'INCLUDE',
    'INCL': 'INCLUDE',
    'DEST': 'SENDTO',
    'DESTINATION': 'SENDTO',
    'RECP': 'SENDTO',
    'RECIPIENT': 'SENDTO',
    'BEG': 'BEGIN',
    'START': 'BEGIN',
    'FROM': 'BEGIN',
}


def unify_options(stream):
    for tok in stream:
        if tok[0] == 'RULE':
            nominal = tok[1]
            yield (tok[0], RULE_MAP.get(nominal, nominal))
        else:
            yield tok


# Parser pass to replace macros with their contents


def define_macros(macros, stream):
    for i in stream:
        if i[0] == 'MACRO':
            name = next(stream, ('EOF', ))
            if name[0] != 'NAME':
                e = "Expected 'NAME' token after MACRO token; got '%s'"
                raise MinderParseError(e % name[0])
            buff = []
            n = next(stream, None)
            while n is not None and n[0] not in SEPS:
                buff.append(n)
                n = next(stream, None)
            macros[name[1]] = buff
            if n is not None and n[0] in SEPS:
                yield from define_macros(macros, chain([n], stream))
        else:
            yield i


def sub_macros(macros, stream):
    for i in stream:
        if i[0] == 'MACROREF':
            macname = i[1][1:]
            if macname not in macros:
                e = 'Reference to undefined macro: %s' % macname
                raise MinderParseError(e)
            else:
                yield from macros[macname]
        else:
            yield i


def handle_macros(macros, stream):
    yield from sub_macros(macros, define_macros(macros, stream))

# Parser pass to unify DATESPEC and TIME tokens

DATESPECS = {'DATE', 'WEEKDAY', 'MONTH', 'WEEKDAY',
             'MONTH', 'MONTHDAY', 'MONTHDAYYEAR',
             'DAY', 'YEAR', 'TIME'}


def merge_two(d1, d2):
    for k in d2:
        if k in d1 and d1[k] != d2[k]:
            if isinstance(d1[k], list):
                d1[k].extend(d2[k])
            else:  # d1[k] is a set
                d1[k].update(d2[k])
        else:
            d1[k] = d2[k]
    return d1


def unify_dates(stream):
    buff = None
    for tok in stream:
        if tok[0] in DATESPECS:
            if buff is None:
                buff = {}
            buff = merge_two(buff, tok[1])
        else:
            if buff is not None:
                yield ('DATESPEC', buff)
                buff = None
            yield tok


# Parser pass to handle OPTIONs

def merge_options(stream):
    for i in stream:
        if i[0] == 'OPTION':
            name = next(stream, ('EOF', ))
            if name[0] not in ('RULE', 'CLASSREF', 'COND'):
                e = "Expected 'RULE' token after OPTION token; got '%s': %s"
                raise MinderParseError(e % (name[0], name, ))
            if name[0] in ('CLASSREF', 'COND'):
                yield ('OPTION', name[0], name[1])
                n = next(stream, None)
                if n is not None and n[0] not in SEPS:
                    e = "Unexpected token after CLASSREF OPTION: %s" % (n, )
                    raise MinderParseError(e)
            else:
                buff = []
                n = next(stream, None)
                while n is not None and n[0] not in SEPS:
                    buff.append(n)
                    n = next(stream, None)
                yield ('OPTION', name[1], buff)
            yield from merge_options(chain([n], stream))
        else:
            yield i

# Parser pass to handle CLASS definitions


def class_definitions(classes, stream):
    for i in stream:
        if i[0] == 'CLASS':
            name = next(stream, ('EOF', ))
            if name[0] != 'NAME':
                e = "Expected 'NAME' token after 'CLASS' token; got '%s'"
                raise MinderParseError(e % name[0])
            n = next(stream, None)
            classes[name[1]] = []
            while n is not None and n[0] == 'OPTION':
                classes[name[1]].append(n)
                n = next(stream, None)
            yield from class_definitions(classes, chain([n], stream))
        else:
            yield i


# extract events from the preprocessed token stream

def add_event(e, l):
    if len(l) > 0 and 'NAME' in {i[1] for i in l[-1]}:
        e.append(sum(l, []))


def define_events(events, levels, markers, stream):
    for i in stream:
        if i[0] == 'EVENT_OR_GROUP':
            add_event(events, levels)
            # set up changes to markers and levels
            marker = i[1]
            if marker in markers:
                # back to a previous indent level
                which = markers.index(marker)
                levels = levels[:which]
                markers = markers[:which]
            elif len(markers) != 0 and len(marker) < len(markers[-1]):
                e = "Event group level (%s) does not match any previous level"
                raise MinderParseError(e.format(i[1]))
            if len(levels) > 0 and 'NAME' in {i[1] for i in levels[-1]}:
                e = "Cannot inherit from named event."
                raise MinderParseError(e)
            if len(levels) == 0:
                levels.append([('OPTION', 'CLASSREF', 'DEFAULT')])
            else:
                levels.append([])
            markers.append(i[1])

            # read in tokens until we hit an OPTION or another EVENT_OR_GROUP
            # these will specify some details (name, time, etc) about the event
            n = next(stream, None)
            while n is not None and n[0] not in ('OPTION', 'EVENT_OR_GROUP'):
                # "up front" details
                if n[0] == 'STRING':
                    levels[-1].append(('OPTION', 'NAME', n[1]))
                elif n[0] == 'CLASSREF':
                    levels[-1].append(('OPTION', 'CLASSREF', n[1]))
                elif n[0] == 'DATESPEC':
                    levels[-1].append(('OPTION', 'DATESPEC', n[1]))
                elif n[0] == 'COND':
                    levels[-1].append(('OPTION', 'COND', n[1]))
                elif n[0] == 'EOF':
                    add_event(events, levels)
                    return events
                else:
                    e = "Unexpected '%s' after EVENT_OR_GROUP: %s"
                    raise MinderParseError(e % (n[0], n))
                n = next(stream, None)

            while n is not None and n[0] == 'OPTION':
                levels[-1].append(n)
                n = next(stream, None)

            if n is not None and n[0] == 'EOF':
                add_event(events, levels)
                return events

            return define_events(events, levels, markers, chain([n], stream))
        elif i[0] == 'EOF':
            add_event(events, levels)
            return events
        else:
            e = "Unexpected Token: %s" % (i, )
            raise MinderParseError(e)


def extract_class_info(c, e):
    o = []
    for i in e:
        _, typ, val = i
        if typ == 'CLASSREF':
            if val not in c:
                raise MinderParseError('Reference to undefined class %s' % val)
            o.extend(c[val])
        else:
            o.append(i)
    return o


def finalize_one_event(c, e):
    out = {'DATESPEC': {}}
    for _, typ, val in extract_class_info(c, e):
        out[typ] = out.get(typ, [])
        if typ == 'DATESPEC':
            out[typ] = merge_two(out[typ], val)
        else:
            out[typ].append(val)
    return out


def finalize_events(classes, events):
    return [finalize_one_event(classes, e) for e in events]


RULE_HANDLERS = {
    'OMIT': rule_omit,
    'REPLACE': rule_replace,
    'INCLUDE': rule_include,
    'REMIND': rule_remind,
    'DURATION': rule_duration,
    'MESSAGE': rule_message,
    'SUBJECT': rule_subject,
    'SENDTO': rule_sendto,
    'FORMAT': rule_format,
    'NAME': rule_name,
}


def rule_pass(events):
    for e in events:
        for (r, f) in RULE_HANDLERS.items():
            e[r] = f(e.get(r, []))
        e['DATESPEC'] = e.get('DATESPEC', {})
        if 'BEGIN' in e:
            e['BEGIN'] = rule_begin(e['BEGIN'])
        if 'END' in e:
            e['END'] = rule_end(e['END'])
        e['DATESPEC'] = [e['DATESPEC']]
        e['DATESPEC'].extend(e['INCLUDE'])
    return events


# End-to-end Parsing


def parse(s, global_macros=None, global_classes=None):
    macros = global_macros if global_macros is not None else {}
    classes = global_classes if global_classes is not None else {}
    tokens = tokenize(s)
    tokens = handle_macros(macros, tokens)
    tokens = unify_options(tokens)
    tokens = unify_dates(tokens)
    tokens = merge_options(tokens)
    tokens = class_definitions(classes, tokens)
    events = define_events([], [], [], tokens)
    events = finalize_events(classes, events)
    events = rule_pass(events)
    return events


def parse_dt_arg(spec, arg, require_time=True):
    t = list(i for i in unify_dates(tokenize(spec)) if i[0] != 'EOF')
    if len(t) > 1 or t[0][0] != 'DATESPEC':
        raise MinderError('%s argument must be a single datespec' % arg)
    t = t[0][1]
    test = is_full_date(t)
    if test is not None:
        raise MinderError(test)
    if require_time and ('time' not in t):
        raise MinderError('%s argument must include a time' % arg)
    return datetime_datespec(t)

