#  Minder: Simple Calendaring System
#  Copyright (C) 2015 Adam Hartz
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Parsing, interpreting, and unifying rules

import re
from datetime import date, datetime

from minder.util import (match_datespec, match_cond, match_range, match_rule,
                         set_peek, is_full_date, apply_change, combine_tdeltas,
                         datetime_datespec, apply_tdelta)


class MinderRuleError(Exception):
    pass


# OMIT


def rule_omit(rules):
    out = []
    for rule in rules:
        try:
            to_loc = rule.index(('KEYWORD', 'TO'))
        except:
            to_loc = -1
        if to_loc < 0:  # just one thing to match
            if len(rule) > 1:
                e = 'Too many values in OMIT rule: %s' % rule
                raise MinderRuleError(e)
            typ, val = rule[0]
            if typ not in ('COND', 'DATESPEC'):
                e = 'OMIT requires COND or DATESPEC, not %s' % typ
                raise MinderRuleError(e)
            out.append(rule[0])
        else:  # check a range
            s = rule[:to_loc]
            e = rule[to_loc + 1:]
            if len(s) != 1:
                e = 'Too many values in OMIT range start: %s' % s
                raise MinderRuleError(e)
            elif len(e) != 1:
                e = 'Too many values in OMIT range end: %s' % s
                raise MinderRuleError(e)
            if s[0][0] != 'DATESPEC' or e[0][0] != 'DATESPEC':
                e = 'OMIT range start and end must be fully-specified dates'
                raise MinderRuleError(e)
            s = s[0][1]
            e = e[0][1]
            se = is_full_date(s)
            ee = is_full_date(e)
            if se is not None:
                raise MinderRuleError('OMIT range start %s' % se)
            if ee is not None:
                raise MinderRuleError('OMIT range end %s' % ee)
            ts1 = list(sorted(s.keys()))
            ts2 = list(sorted(e.keys()))
            if ts1 != ts2:
                e = 'OMIT range start and end not the same type'
                raise MinderRuleError(e)
            sd = datetime_datespec(s)
            ed = datetime_datespec(e)
            out.append(('RANGE', (sd, ed)))
    return out


def handle_omit(d, rules):
    for rule in rules:
        if match_rule(d, rule):
            return True
    return False

# REPLACE


def unify_reltime(rule):
    buff = None
    for tok in rule:
        if tok[0] == 'RELTIME':
            if buff is None:
                buff = {}
            buff = combine_tdeltas(buff, tok[1])
        else:
            if buff is not None:
                yield ('RELTIME', buff)
                buff = None
            yield tok
    if buff is not None:
        yield ('RELTIME', buff)


def rule_replace(rules):
    out = []
    for rule in rules:
        rule = list(unify_reltime(rule))
        try:
            with_loc = rule.index(('KEYWORD', 'WITH'))
        except:
            with_loc = -1
        if with_loc == 0:
            test = ('COND', 'True')
        elif with_loc == 1:
            test = rule[0]
        else:
            e = "REPLACE rule must contain 'WITH' in position 0 or 1"
            raise MinderRuleError(e)
        post = rule[with_loc + 1:]
        if len(post) != 1:
            e = "REPLACE rule must contain exactly one element after WITH'"
            raise MinderRuleError(e)
        post = post[0]
        if post[0] not in ('RELTIME'):
            e = "REPLACE rule after WITH must be RELTIME"
            raise MinderRuleError(e)
        typ, val = test
        if typ not in ('DATESPEC', 'COND'):
            e = 'REPLACE requires COND or DATESPEC, not %s' % typ
            raise MinderRuleError(e)
        out.append((test, post))
    return out


# BEGIN


def rule_begin(rules):
    if len(rules) > 1:
        raise MinderRuleError('An event may have at most one BEGIN rule.')
    if len(rules) == 0:
        return date(MINYEAR, 1, 1)
    rule = rules[0]
    if len(rule) != 1:
        raise MinderRuleError('An event may have at most one BEGIN rule.')
    typ, val = rule[0]
    if typ != 'DATESPEC' or (is_full_date(val) is not None):
        raise MinderRuleError('BEGIN expects a full DATESPEC')
    return datetime_datespec(val)


# END


def rule_end(rules):
    if len(rules) > 1:
        raise MinderRuleError('An event may have at most one END rule.')
    if len(rules) == 0:
        return date(MAXYEAR, 12, 31)
    rule = rules[0]
    if len(rule) != 1:
        raise MinderRuleError('An event may have at most one END rule.')
    typ, val = rule[0]
    if typ != 'DATESPEC' or (is_full_date(val) is not None):
        raise MinderRuleError('END expects a full DATESPEC')
    return datetime_datespec(val)


# INCLUDE


def rule_include(rules):
    out = []
    for rule in rules:
        if len(rule) != 1 or rule[0] != 'DATESPEC':
            raise MinderRuleError('INCLUDE expects exactly one DATESPEC')
        out.append(rule)
    return out


# REMIND


def rule_remind(rules):
    out = []
    for rule in rules:
        for tok in rule:
            if tok[0] not in {'RELTIME', 'TDELTA'}:
                e = 'REMIND options must be RELTIME or TDELTA'
                raise MinderRuleError(e)
            elif tok[0] == 'TDELTA':
                new = tok[1]
                new['neg'] = True
            elif tok[0] == 'RELTIME':
                if not tok[1]['neg']:
                    e = 'REMIND option RELTIME must be negative'
                    raise MinderRuleError(e)
                new = tok[1]
            out.append(new)
    return out


# DURATION


def rule_duration(rules):
    out = None
    for rule in rules:
        if len(rule) > 1 or rule[0][0] != 'TDELTA':
            e = 'DURATION rule must contain exactly one TDELTA'
            raise MinderRuleError(e)
        out = rule[0][1]
    return out


# MESSAGE


def rule_message(rules):
    out = None
    for rule in rules:
        if len(rule) > 1 or rule[0][0] != 'STRING':
            e = 'MESSAGE rule must contain exactly one STRING'
            raise MinderRuleError(e)
        out = rule[0][1]
    return out


# SUBJECT


def rule_subject(rules):
    out = None
    for rule in rules:
        if len(rule) > 1 or rule[0][0] != 'STRING':
            e = 'SUBJECT rule must contain exactly one STRING'
            raise MinderRuleError(e)
        out = rule[0][1]
    return out


# SENDTO

EMAIL_REGEX = re.compile('^[^@]+@[^@]+\.[^@]+$')


def rule_sendto(rules):
    out = []
    for rule in rules:
        for tok in rule:
            if tok[0] != 'STRING' or (not EMAIL_REGEX.match(tok[1])):
                e = ('SENDTO arguments must be '
                     'strings containing e-mail addresses')
                raise MinderRuleError(e)
            out.append(tok[1])
    return out


# FORMAT


def rule_format(rules):
    out = None
    for rule in rules:
        if len(rule) > 1 or rule[0][0] != 'STRING':
            e = 'FORMAT rule must contain exactly one STRING'
            raise MinderRuleError(e)
        out = rule[0][1]
    return out


# NAME


def rule_name(rules):
    return rules[-1]
