#  Minder: Simple Calendaring System
#  Copyright (C) 2015 Adam Hartz
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Utilities for handling triggers

from datetime import date, time, datetime, timedelta, MAXYEAR

from minder.util import *
from minder.rule import *


# Handling times in triggers


def _time(t):
    return time(hour=t[0], minute=t[1])


def _with_time(tfunc):
    def _wt_helper(trigger, start, max_date):
        n = tfunc(trigger, start, max_date)
        d = start.date()
        if n is None:
            return None
        elif n == d:
            nt = start.time()
            for t in sorted(map(_time, trigger['time'])):
                if nt <= t:
                    return datetime.combine(n, t)
            else:
                # past all the times on this day; use midnight the following
                # day as our new starting point
                start = datetime.combine(d+timedelta(days=1), time(0, 0))
                return _wt_helper(trigger, start, max_date)
        elif n > d:
            return datetime.combine(n, _time(min(trigger['time'])))
        else:
            return None
    return _wt_helper


# Trigger handlers (date-level)


def next_trigger_none(trigger, start, max_date):
    return start.date()


def next_trigger_day(trigger, start, max_date):
    while lte(start, max_date):
        for d in sorted(trigger['day']):
            if start.day <= d and valid_date(start.year, start.month, d):
                return date(start.year, start.month, d)
        days_left = days_in_month(start.year, start.month) - start.day
        start += timedelta(days=days_left + 1)


def next_trigger_month(trigger, start, max_date):
    for m in sorted(trigger['month']):
        if start.month == m:
            return start.date()
        if start.month < m:
            return date(start.year, m, 1)
    return date(start.year + 1, min(trigger['month']), 1)


def next_trigger_year(trigger, start, max_date):
    for y in sorted(trigger['year']):
        if start.year == y:
            return start.date()
        if start.year < y:
            return date(y, 1, 1)
    return None


def next_trigger_day_month(trigger, start, max_date):
    x = {(m, d) for m in trigger['month']
         for d in trigger['day']
         if 1 <= d <= MONTH_DAYS[m]}
    for (m, d) in sorted(x):
        if start.month > m:
            continue
        if start.month == m and start.day > d:
            continue
        if not valid_date(start.year, m, d):
            continue
        return date(start.year, m, d)
    # no date found; start looping
    y = start.year + 1
    while lte(start, max_date):
        for (m, d) in sorted(x):
            if valid_date(y, m, d):
                return date(y, m, d)
        y += 1


def next_trigger_day_year(trigger, start, max_date):
    for y in sorted(trigger['year']):
        for m in range(1, 13):
            for d in sorted(trigger['day']):
                if valid_date(y, m, d):
                    d = date(y, m, d)
                    if d > start.date():
                        return d
    return None


def next_trigger_month_year(trigger, start, max_date):
    for y in sorted(trigger['year']):
        if y < start.year:
            continue
        for m in sorted(trigger['month']):
            if y == start.year:
                if m == start.month:
                    return start.date()
                elif m < start.month:
                    continue
                else:
                    return date(y, m, 1)
            else:
                return date(y, m, 1)
    return None


def next_trigger_day_month_year(trigger, start, max_date):
    sd = start.date()
    for y in sorted(trigger['year']):
        for m in sorted(trigger['month']):
            for d in sorted(trigger['day']):
                if valid_date(y, m, d):
                    dt = date(y, m, d)
                    if dt >= sd:
                        return dt
    return None


def _add_dow_check(fn):
    def _with_dow(trigger, start, max_date):
        sd = start
        while lte(sd, max_date):
            d = fn(trigger, sd, max_date)
            if d is None:
                return
            if d.weekday() in trigger['dow']:
                return d
            sd = datetime.combine(d, time(0, 0)) + timedelta(days=1)
    return _with_dow


TRIGGER_FUNCTIONS = {
    tuple(): next_trigger_none,
    ('day', ): next_trigger_day,
    ('day', 'month'): next_trigger_day_month,
    ('day', 'month', 'year'): next_trigger_day_month_year,
    ('day', 'year'): next_trigger_day_year,
    ('month', ): next_trigger_month,
    ('month', 'year'): next_trigger_month_year,
    ('year', ): next_trigger_year,
}

for k, v in list(TRIGGER_FUNCTIONS.items()):
    TRIGGER_FUNCTIONS[tuple(sorted(k + ('dow', )))] = _add_dow_check(v)

for k, v in list(TRIGGER_FUNCTIONS.items()):
    TRIGGER_FUNCTIONS[tuple(sorted(k + ('time', )))] = _with_time(v)


def get_next_simple_trigger(ds, start, max_date):
    k = tuple(sorted(ds.keys()))
    f = TRIGGER_FUNCTIONS[k]
    return f(ds, start, max_date)


def get_next_replace_trigger(event, start, max_date):
    max_replace_tdelta = timedelta()
    d = start
    for r in event['REPLACE']:
        d_new = apply_tdelta(d, r[1][1])
        delta = abs(d_new - start)
        if delta > max_replace_tdelta:
            max_replace_tdelta = delta
    max_replace_tdelta += timedelta(minutes=1)
    max_alert_time = None
    dt = datetime.now()
    orig_start = start
    for a in event['REMIND']:
        d = apply_tdelta(dt, a) - dt
        if max_alert_time is None or d < max_alert_time:
            max_alert_time = abs(d)
    try:
        md = max_date + max_replace_tdelta + (max_alert_time or timedelta())
    except:
        md = max_date
    start = start - max_replace_tdelta
    n = None
    e = dict(event)
    del e['REPLACE']
    all_trigs = get_triggers_in_range(e, start, md)
    for t in all_trigs:
        for rule, change in event['REPLACE']:
            if match_rule(t, rule):
                real = apply_tdelta(t, change[1])
            else:
                real = t
            if (n is None or real < n) and (gte(real, orig_start) and lte(real, max_date)):
                n = real
    return n


def get_next_trigger(event, start, max_date=datetime(MAXYEAR, 12, 30)):
    event = dict(event)
    # set start of search
    b = event.get('BEGIN', start)
    if gt(b, start):
        start = b
    if not isinstance(start, datetime):
        start = datetime.combine(start, time(0, 0))
    # set end of search
    e = event.get('END', max_date)
    if lt(e, max_date):
        max_date = e
    # if this has a REPLACE rule, we will need to do something
    # completely different...
    if 'REPLACE' in event and len(event['REPLACE']) > 0:
        return get_next_replace_trigger(event, start, max_date)
    # find next simple trigger
    nexttrig = min_date([get_next_simple_trigger(d, start, max_date)
                         for d in event['DATESPEC']])
    if nexttrig is None:
        return None
    # if needs to be omitted, find next trigger...
    if handle_omit(nexttrig, event['OMIT']):
        if isinstance(nexttrig, datetime):
            nexttrig += timedelta(minutes=1)
        else:
            nexttrig += timedelta(days=1)
        return get_next_trigger(event, nexttrig, max_date)
    # if we're here, return the next trigger
    return nexttrig


def get_triggers_in_range(event,
                          start,
                          max_date=datetime(MAXYEAR, 12, 30)):
    out = []
    event = dict(event)
    t = get_next_trigger(event, start, max_date)
    while t is not None and lte(t, max_date):
        out.append(t)
        if isinstance(t, datetime):
            start = t + timedelta(minutes=1)
        else:
            start = t + timedelta(days=1)
        t = get_next_trigger(event, start, max_date)
    return out


def get_triggers_could_need_alert(event,
                                  start,
                                  max_date=datetime(MAXYEAR, 12, 30)):
    event = dict(event)
    out = []
    max_alert_time = None
    dt = datetime.now()
    orig_start = start
    for a in event['REMIND']:
        d = apply_tdelta(dt, a) - dt
        if max_alert_time is None or d < max_alert_time:
            max_alert_time = d
    if max_alert_time is None:
        return out
    t = get_next_trigger(event, start, max_date)
    if t is not None:
        out.append(t)
    last = t
    count = 0
    while t is not None and (orig_start - t) >= max_alert_time:
        if t == last:
            count += 1
        else:
            count = 0
            out.append(t)
            last = t
        if isinstance(t, datetime):
            start = t + timedelta(minutes=count+1)
        else:
            start = t + timedelta(days=count+1)
        t = get_next_trigger(event, start, max_date)
    return out
