#  Minder: Simple Calendaring System
#  Copyright (C) 2015 Adam Hartz
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# calendar utils

import os
import re

import time as pytime
from datetime import date, time, datetime, timedelta


class MinderError(Exception):
    pass


class MinderDateError(MinderError):
    pass


class MinderTimeError(MinderError):
    pass


MONTH_DAYS = [None, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]


def days_in_month(year, month):
    ny = year
    nm = month + 1
    if nm > 12:
        nm = 1
        ny += 1
    return (datetime(ny, nm, 1) - datetime(year, month, 1)).days


def valid_date(year, month, date):
    return 1 <= date <= days_in_month(year, month)


def jump_month(dt, frac=1):
    return dt + timedelta(days=frac * days_in_month(dt.year, dt.month))


def backward_jump_month(dt, frac=1):
    y = dt.year
    m = dt.month - 1
    if m == 0:
        m = 12
        y -= 1
    return dt - timedelta(days=frac * days_in_month(y, m))


def jump_months(dt, n):
    if n < 0:
        n = -n
        f = backward_jump_month
    else:
        f = jump_month
    for i in range(int(n)):
        dt = f(dt)
    dt = f(dt, frac=n % 1.0)
    return dt


def jump_years(dt, n):
    return jump_months(dt, 12*n)


_dt_matchers = {
    'time': lambda d: (d.time().hour, d.time().minute),
    'day': lambda d: d.day,
    'month': lambda d: d.month,
    'year': lambda d: d.year,
    'dow': lambda d: d.weekday()
}


def match_datespec(d, datespec):
    isdt = isinstance(d, datetime)
    hastime = 'time' in datespec
    if hastime and not isdt:
        return False
    for k, v in datespec.items():
        if _dt_matchers[k](d) not in v:
            return False
    return True


def match_cond(d, cond):
    e = {k: v(d) for k, v in _dt_matchers.items() if k != 'time'}
    if isinstance(d, datetime):
        e['time'] = _dt_matchers['time'](d)
    e['match'] = d
    return bool(eval(cond, e))


def match_range(d, rnge):
    start, end = rnge
    return lte(start, d) and lte(d, end)


def match_rule(d, r):
    typ, val = r
    if typ == 'COND':
        v = match_cond(d, val)
    elif typ == 'DATESPEC':
        v = match_datespec(d, val)
    elif typ == 'RANGE':
        v = match_range(d, val)
    else:
        v = False
    return v


def is_unambiguous(d):
    if any(len(i) != 1 for i in d.values()):
        return 'must be unambiguous'


def is_full_date(d):
    if any(i not in d for i in ('day', 'month', 'year')):
        return 'must be a fully-specified date or datetime'
    if 'dow' in d:
        return 'may not contain "day of week" condition'
    return is_unambiguous(d)


def combine_tdeltas(td1, td2):
    mul = -1 if td2.get('neg', False) else 1
    sig = -1 if td1.get('neg', False) else 1
    for k, v in td2.items():
        if k == 'neg':
            continue
        td1[k] = td1.get(k, 0) + sig * mul * v
    return td1


def apply_tdelta(d, delta):
    y = delta.get('years', 0)
    m = delta.get('months', 0)
    fac = -1 if delta.get('neg', False) else 1
    d = jump_months(jump_years(d, fac*y), fac*m)
    args = {k: v
            for k, v in delta.items()
            if k in ('days', 'hours', 'minutes')}
    return d + fac * timedelta(**args)


def apply_change(d, c):
    if c[0] == 'RELTIME':
        return apply_tdelta(d, c[1])
    elif c[0] == 'DATESPEC':
        return apply_datespec(d, c[1])


def set_peek(s):
    for x in s:
        break
    return x


def datetime_datespec(d):
    nd = date(set_peek(d['year']), set_peek(d['month']), set_peek(d['day']))
    if 'time' in d:
        t = set_peek(d['time'])
        nd = datetime.combine(nd, time(*t))
    return nd


def compare(f):
    def cmp(d1, d2):
        t1 = isinstance(d1, datetime)
        t2 = isinstance(d2, datetime)
        if t1 and not t2:
            return f(d1.date(), d2)
        elif t2 and not t1:
            return f(d1, d2.date())
        else:
            return f(d1, d2)
    return cmp

gt = compare(lambda x, y: x > y)
lt = compare(lambda x, y: x < y)
gte = compare(lambda x, y: x >= y)
lte = compare(lambda x, y: x <= y)
eq = compare(lambda x, y: x == y)
ne = compare(lambda x, y: x != y)


def min_date(ds):
    best = None
    for d in ds:
        if best is None:
            best = d
            continue
        if d is None:
            continue
        if lt(d, best):
            best = d
    return best


def from_timestamp(x):
    return datetime.strptime(x, '%Y-%m-%d:%H:%M:%S.%f')


def timestamp(x):
    return x.strftime('%Y-%m-%d:%H:%M:%S.%f')


def unix_timestamp(dt):
    return str(pytime.mktime(dt.timetuple()))


FLOAT_REGEX = re.compile(r'^(\d+\.\d*)|(\.\d+)$')


def from_unix_timestamp(ut):
    assert FLOAT_REGEX.match(ut) is not None
    return datetime.fromtimestamp(eval(ut))


def ensure_dirs(path):
    if not os.path.isdir(path):
        os.makedirs(path)

CONFIG_LOCATION = os.path.expanduser(os.path.join('~', '.config', 'minder'))
ensure_dirs(CONFIG_LOCATION)


def all_calendars(filelist, recurse=False):
    while not all(i.endswith('.mnd') and os.path.isfile(i) for i in filelist):
        to_remove = set()
        replace = {}
        for ix, f in enumerate(filelist):
            if os.path.isdir(f):
                if recurse:
                    replace[ix] = [os.path.join(f, i)
                                   for i in os.listdir(f)]
                else:
                    to_remove.add(ix)
            elif os.path.isfile(f):
                if not f.endswith('.mnd'):
                    to_remove.add(ix)
            else:
                to_remove.add(ix)
        keys = to_remove.union(set(replace.keys()))
        for ix in sorted(keys, reverse=True):
            if ix in to_remove:
                filelist = filelist[:ix] + filelist[ix + 1:]
            else:
                filelist = filelist[:ix] + replace[ix] + filelist[ix + 1:]
    return filelist
